/*
 This file is part of the digitalSTROM dsm-api library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM dsm-api library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DSM_MACROS_H_
#define DSM_MACROS_H_

#define RESPONSE_DATA_OFFSET 3
#define REQUEST_DATA_OFFSET 2
#define COMMAND_OFFSET 0
#define MODIFIER_OFFSET 1

#define GetCommand(msg_) \
    (msg_).data[COMMAND_OFFSET]

#define SetCommand(msg_, command_) \
    (msg_).data[COMMAND_OFFSET] = (command_)

#define GetModifier(msg_) \
    (msg_).data[MODIFIER_OFFSET]

#define SetModifier(msg_, command_) \
    (msg_).data[MODIFIER_OFFSET] = (command_)

#define ResponseGetData8(msg_, offset_) \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)]

#define ResponseGetData16(msg_, offset_) \
    ((uint16_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+0]) << 8) | \
    ((uint16_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+1]) << 0)

#define ResponseGetData32(msg_, offset_) \
    ((uint32_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+0]) << 24) | \
    ((uint32_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+1]) << 16) | \
    ((uint32_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+2]) <<  8) | \
    ((uint32_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+3]) <<  0)

#define ResponseGetData64(msg_, offset_) \
    ((uint64_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+0]) << 56) | \
    ((uint64_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+1]) << 48) | \
    ((uint64_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+2]) << 40) | \
    ((uint64_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+3]) << 32) | \
    ((uint64_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+4]) << 24) | \
    ((uint64_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+5]) << 16) | \
    ((uint64_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+6]) <<  8) | \
    ((uint64_t)((msg_).data[RESPONSE_DATA_OFFSET+(offset_)+7]) <<  0)

#define ResponseGetArray8(msg_, offset_, array_, size_) \
    memcpy((array_), &((msg_).data[RESPONSE_DATA_OFFSET+(offset_)]), size_)

#define RequestGetData8(msg_, offset_) \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)]

#define RequestGetData16(msg_, offset_) \
    ((uint16_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+0]) << 8) | \
    ((uint16_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+1]) << 0)

#define RequestGetData32(msg_, offset_) \
    ((uint32_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+0]) << 24) | \
    ((uint32_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+1]) << 16) | \
    ((uint32_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+2]) <<  8) | \
    ((uint32_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+3]) <<  0)

#define RequestGetData64(msg_, offset_) \
    ((uint64_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+0]) << 56) | \
    ((uint64_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+1]) << 48) | \
    ((uint64_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+2]) << 40) | \
    ((uint64_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+3]) << 32) | \
    ((uint64_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+4]) << 24) | \
    ((uint64_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+5]) << 16) | \
    ((uint64_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+6]) <<  8) | \
    ((uint64_t)((msg_).data[REQUEST_DATA_OFFSET+(offset_)+7]) <<  0)

#define RequestGetArray8(msg_, offset_, array_, size_) \
    memcpy((array_), &((msg_).data[REQUEST_DATA_OFFSET+(offset_)]), size_)

#define ResponseSetData8(msg_, offset_, data8_) do { \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)] = (data8_); \
} while (0)

#define ResponseSetData16(msg_, offset_, data16_) do { \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+0] = (data16_) >> 8; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+1] = (data16_) >> 0; \
} while (0)

#define ResponseSetData32(msg_, offset_, data32_) do { \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+0] = (data32_) >> 24; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+1] = (data32_) >> 16; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+2] = (data32_) >>  8; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+3] = (data32_) >>  0; \
} while (0)

#define ResponseSetData64(msg_, offset_, data64_) do { \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+0] = (data64_) >> 56; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+1] = (data64_) >> 48; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+2] = (data64_) >> 40; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+3] = (data64_) >> 32; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+4] = (data64_) >> 24; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+5] = (data64_) >> 16; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+6] = (data64_) >>  8; \
    (msg_).data[RESPONSE_DATA_OFFSET+(offset_)+7] = (data64_) >>  0; \
} while (0)

#define ResponseSetArray8(msg_, offset_, array_, size_) do { \
    memcpy(&((msg_).data[REQUEST_DATA_OFFSET+(offset_)]), (array_), size_); \
} while (0)

#define RequestSetData8(msg_, offset_, data8_) do { \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)] = (data8_); \
} while (0)

#define RequestSetData16(msg_, offset_, data16_) do { \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+0] = (data16_) >> 8; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+1] = (data16_) >> 0; \
} while (0)

#define RequestSetData32(msg_, offset_, data32_) do { \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+0] = (data32_) >> 24; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+1] = (data32_) >> 16; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+2] = (data32_) >>  8; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+3] = (data32_) >>  0; \
} while (0)

#define RequestSetData64(msg_, offset_, data64_) do { \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+0] = (data64_) >> 56; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+1] = (data64_) >> 48; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+2] = (data64_) >> 40; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+3] = (data64_) >> 32; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+4] = (data64_) >> 24; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+5] = (data64_) >> 16; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+6] = (data64_) >>  8; \
    (msg_).data[REQUEST_DATA_OFFSET+(offset_)+7] = (data64_) >>  0; \
} while (0)

#define RequestSetArray8(msg_, offset_, array_, size_) do { \
    memcpy(&((msg_).data[REQUEST_DATA_OFFSET+(offset_)]), (array_), size_); \
} while (0)

#define RequestSetStringArray8(msg_, offset_, string_, size_) do { \
    strncpy(&((msg_).data[REQUEST_DATA_OFFSET+(offset_)]), (string_), size_); \
} while (0)

#define SetParameterLength(msg_, count_) \
    (msg_).length = (count_) + 2

#define SetDSUID(dst_, src_) \
    memcpy(&(dst_), &(src_), sizeof(dsuid_t))

#define GetErrorCode(msg_) \
    ResponseGetData8(msg_, -1)

#define GetErrorCode(msg_) \
    ResponseGetData8(msg_, -1)

#endif /* DSM_MACROS_H_ */
