# This file is part of the digitalSTROM dsm-api library.
#
# Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland
#
# digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# digitalSTROM dsm-api library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.
 
from string import Template
from Model import *
#extend class Parameter with methods for C generation

def getDoc(self, isReturn=False):
    out = "@param ["
    if (isReturn):
        out += "out"
    else:
        out += "in"
    out += "] " + self.name + " " + self.description
    if (isReturn and self.type_count > 1):
        out += " (pass a reference to a preallocated buffer of at least " + str(self.type_count) + " bytes.)"
    return out
Parameter.getDoc = getDoc

def signature(self):
    if self.name == "DSID" and self.type_size == 8 and self.type_count == 12:
        return "dsid_t DSID"
    elif self.name == "dSUID" and self.type_size == 8 and self.type_count == 17:
        return "dsuid_t dSUID"
    else:
        if (self.type_count > 1):
            out = "const uint" + str(self.type_size) + "_t *"
        else:
            out = "uint" + str(self.type_size) + "_t "
        out += self.name
        return out
Parameter.signature = signature

def declare_by_reference(self):
    if self.name == "DSID" and self.type_size == 8 and self.type_count == 12:
        return "dsid_t *DSID"
    elif self.name == "dSUID" and self.type_size == 8 and self.type_count == 17:
        return "dsuid_t *dSUID"
    else:
        return "uint" + str(self.type_size) + "_t *" + self.name
Parameter.declare_by_reference = declare_by_reference

def getFromResponse(self):
    out = "    if (" + self.name + " != NULL)\n"
    out += "    {\n"
    if self.name == "DSID" and self.type_size == 8 and self.type_count == 12:
        out += "        ResponseGetArray" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ", DSID->id, " + str(self.type_count) + ");\n"
    elif self.name == "dSUID" and self.type_size == 8 and self.type_count == 17:
        out += "        ResponseGetArray" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ", dSUID->id, " + str(self.type_count) + ");\n"
        if self.add_header:
            out += """        if ((VendorId != NULL) && (ProductId != NULL) && (FunctionId != NULL) && (Version != NULL)) {
            DsmApiAddDeviceDSUIDHeader(*VendorId, *ProductId, (*FunctionId >> 12) & 0xf, *Version, dSUID);
        }\n"""
    elif (self.type_count == 1):
        out += "        *" + self.name + " = ResponseGetData" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ");\n"
    elif (self.dynamic):
        out += "        ResponseGetArray" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ", " + self.name + ", (*" + self.getDynamicLength() + " < " + str(self.type_count) + ") ? *" + self.getDynamicLength() + " : " + str(self.type_count) + ");\n"
    else:
        out += "        ResponseGetArray" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ", " + self.name + ", " + str(self.type_count) + ");\n"
    out += "    }\n"
    return out
Parameter.getFromResponse = getFromResponse

def getFromRequest(self):
    out = "    if (" + self.name + " != NULL)\n"
    out += "    {\n"
    if self.name == "DSID" and self.type_size == 8 and self.type_count == 12:
        out += "        RequestGetArray" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ", DSID->id, " + str(self.type_count) + ");\n"
    elif self.name == "dSUID" and self.type_size == 8 and self.type_count == 17:
        out += "        RequestGetArray" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ", dSUID->id, " + str(self.type_count) + ");\n"
    elif (self.type_count == 1):
        out += "        *" + self.name + " = RequestGetData" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ");\n"
    elif (self.dynamic):
        out += "        RequestGetArray" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ", " + self.name + ", (*" + self.getDynamicLength() + " < " + str(self.type_count) + ") ? *" + self.getDynamicLength() + " : " + str(self.type_count) + ");\n"
    else:
        out += "        RequestGetArray" + str(self.type_size) + "(*pFrameAnswer, " + str(self.offset) + ", " + self.name + ", " + str(self.type_count) + ");\n"
    out += "    }\n"
    return out
Parameter.getFromRequest = getFromRequest

def setRequest(self):
    if self.name == "DSID" and self.type_size == 8 and self.type_count == 12:
        return "RequestSetArray" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", DSID.id, " + str(self.type_count) + ");"
    elif self.name == "dSUID" and self.type_size == 8 and self.type_count == 17:
        return "RequestSetArray" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", dSUID.id, " + str(self.type_count) + ");"
    elif self.name == "Name" and self.type_size == 8 and self.type_count == 21:
        return "RequestSetStringArray" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", " + self.name + ", " + str(self.type_count) + ");"
    elif (self.type_count == 1):
        return "RequestSetData" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", " + self.name + ");"
    elif (self.dynamic):
        return "RequestSetArray" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", " + self.name + ", (" + self.getDynamicLength() + " < " + str(self.type_count) + ") ? " + self.getDynamicLength() + " : " + str(self.type_count) + ");"
    else:
        return "RequestSetArray" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", " + self.name + ", " + str(self.type_count) + ");"
Parameter.setRequest = setRequest

def setResponse(self):
    if self.name == "DSID" and self.type_size == 8 and self.type_count == 12:
        return "ResponseSetData" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", DSID.id, " + str(self.type_count) + ");"
    elif self.name == "dSUID" and self.type_size == 8 and self.type_count == 17:
        return "ResponseSetData" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", dSUID.id, " + str(self.type_count) + ");"
    elif (self.type_count == 1):
        return "ResponseSetData" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", " + self.name + ");"
    elif (self.dynamic):
        return "ResponseSetArray" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", " + self.name + ", (" + self.getDynamicLength() + " < " + str(self.type_count) + ") ? " + self.getDynamicLength() + " : " + str(self.type_count) + ");"
    else:
        return "ResponseSetArray" + str(self.type_size) + "(pFrame, " + str(self.offset) + ", " + self.name + ", " + str(self.type_count) + ");"
Parameter.setResponse = setResponse

def call_as_object_by_reference(self):
    out = ""
    if (self.type_count == 1) or \
       (self.name == "DSID" and self.type_size == 8 and self.type_count == 12) or \
       (self.name == "dSUID" and self.type_size == 8 and self.type_count == 17):
        out += "&"
    out += self.name
    return out
Parameter.call_as_object_by_reference = call_as_object_by_reference

def call(self):
    return self.name
Parameter.call = call

def declaration(self):
    if self.name == "DSID" and self.type_size == 8 and self.type_count == 12:
        return "dsid_t DSID"
    elif self.name == "dSUID" and self.type_size == 8 and self.type_count == 17:
        return "dsuid_t dSUID"
    else:
        out = "uint" + str(self.type_size) + "_t " + self.name
        if (self.type_count > 1):
            out += "[" + str(self.type_count) + "]"
        return out
Parameter.declaration = declaration
    

def documentation_call(self):
    s = Template("""/**
 * $description
 * $deprecated
 * $vdsm_only
 * This call has a timeout of $timeout seconds.
 *
 * To receive asynchronous responses to this call (e.g. when addressing broadcast) register the #$response_callback callback.
 * To receive this request from other bus members register the #$request_callback callback.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dS485DeviceUID Destination device of the call
 * $request_response
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */""")
    return s.substitute(deprecated = "@deprecated This function is deprecated!\n" if self.deprecated else "",
                        vdsm_only = "@vdsm_only This function can be used only within a vDsm!\n" if self.long_message else "",
                        description = self.description,
                        timeout = self.timeout,
                        response_callback = self.name + "_response_callback_t",
                        request_callback = self.name + "_request_callback_t",
                        request_response = "\n * ".join(map(lambda x: x.getDoc(), self.request_params) + 
                                                        map(lambda x: x.getDoc(True), self.response_params)))
ApiCall.documentation_call = documentation_call

def call_signature(self, header=False):
    out = "int " + self.name + "(DsmApiHandle_t handle, dsuid_t dS485DeviceUID"
    if self.request_params:
        out += ", " + ", ".join(map(lambda x: x.signature(), self.request_params))
    if self.response_params:
        out += ", " + ", ".join(map(lambda x: x.declare_by_reference(), self.response_params))
    out += ")"
    if header and self.deprecated:
        out += "\n#ifndef SWIGLUA\n__attribute__ ((deprecated))\n#endif\n"
    return out
ApiCall.call_signature = call_signature

def callback_documentation(self, call_type):
    s = Template("""/**
 * $description
 * $deprecated
 * $vdsm_only
 * $type callback function prototype${call}.
 * 
 * To register this callback function, call the #DsmApiSetCallback function
 * with MessageType = #${message_type}, MessageId = #$messageId and ModifierId = #$modifierId
 *
 * @param [in] error Error result of the (potential) previous call.
 * @param [in] arg User data
 * @param [in] sourceId Source address of the call
 * @param [in] destinationId Destination address of the call
 * $parameters
 */""")
    if call_type == "response":
        type_const = "DS485_CONTAINER_RESPONSE"
        parameters = self.response_params
    elif call_type == "request":
        type_const = "DS485_CONTAINER_REQUEST"
        parameters = self.request_params
    else:
        type_const = "DS485_CONTAINER_EVENT"
        parameters = self.response_params
        
    return s.substitute(deprecated = "@deprecated This function is deprecated!\n" if self.deprecated else "",
                        vdsm_only = "@vdsm_only This function can be used only within a vDsm!\n" if self.long_message else "",
                        description = self.description,
                        type = call_type.capitalize(),
                        call = " for call #" + self.name if call_type != "event" else "",
                        message_type = type_const,
                        parameters = "\n * ".join(map(lambda x: x.getDoc(), parameters)), 
                        messageId = self.message_id,
                        modifierId = self.modifier_id)
ApiCall.callback_documentation = callback_documentation

def callback_declaration(self, call_type):
    if call_type == "request":
        parameters = self.request_params
    else:
        parameters = self.response_params
    out = "typedef void (*" + self.name + "_" + call_type + "_callback_t)(uint8_t error_code, void *arg, dsuid_t sourceId, dsuid_t destinationId"
    if parameters:
        out += ", " + ", ".join(map(lambda x: x.signature(), parameters))
    out += ")"
    if self.deprecated:
        out += "\n#ifndef SWIGLUA\n__attribute__ ((deprecated))\n#endif\n"
    out += ";"
    return out
ApiCall.callback_declaration = callback_declaration

def parse_method_signature(self, call_type):
    if call_type == "request":
        parameters = self.request_params
    else:
        parameters = self.response_params
    
    if not self.long_message:
        out = "static int " + self.name + "_parse_" + call_type + "(const ds485_container_t *pFrameAnswer"
    else:
        out = "static int " + self.name + "_parse_" + call_type + "(const ds485n_packet_t *pFrameAnswer"
    if parameters:
        out += ", " + ", ".join(map(lambda x: x.declare_by_reference(), parameters))
    out += ")"
    return out
ApiCall.parse_method_signature = parse_method_signature

def parse_method(self, call_type):
    out = self.parse_method_signature(call_type) + "\n"
    out += "{\n"
    s = Template("""    if (GetCommand(*pFrameAnswer) != $messageId) {
        return ERROR_REC_WRONG_FUNCTION_ID;
    }
    if (GetModifier(*pFrameAnswer) != $modifierId) {
        return ERROR_REC_WRONG_MODIFIER;
    }
""")
    out += s.substitute(messageId = self.message_id,
                       modifierId = self.modifier_id)
    if call_type == "request":
        out += "".join(map(lambda x: x.getFromRequest(), self.request_params))
        out += "    return 0;\n"
    else:
        out += "".join(map(lambda x: x.getFromResponse(), self.response_params))
        out += "    return GetErrorCode(*pFrameAnswer);\n"
    out += "}"
    return out
ApiCall.parse_method = parse_method

def parse_method_call(self, call_type):
    if call_type == "request":
        parameters = self.request_params
    else:
        parameters = self.response_params
    
    out = self.name + "_parse_" + call_type + "(&pFrameAnswer"
    if parameters:
        out += ", " + ", ".join(map(lambda x: x.call(), parameters))
    out += ")"
    return out
ApiCall.parse_method_call = parse_method_call

def sync_method(self):
    s = Template("""$signature
{
    $container_type pFrame;
    $container_type pFrameAnswer;
    dsm_api_data_t *data = (dsm_api_data_t*)handle;
    int result;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    memset(&pFrame, 0, sizeof($container_type));
    pFrame.containerType = DS485_CONTAINER_REQUEST;
    SetDSUID(pFrame.destinationId, dS485DeviceUID);
    SetCommand(pFrame, $message_id);
    SetModifier(pFrame, $modifier_id);
    $parameter
    SetParameterLength(pFrame, $parameter_length);

    result = $send_function(data->client, &pFrame, &pFrameAnswer, $timeout);
    if (result == 0)
    {
        return $parse_response;
    }
    else
    {
        return result;
    }
}""")
    return s.substitute(signature = self.call_signature(),
                        container_type = "ds485_container_t" if not self.long_message else "ds485n_packet_t",
                        send_function = "ds485_client_send_sync_command" if not self.long_message else "ds485_client_send_sync_long_command",
                        message_id = self.message_id,
                        modifier_id = str(self.modifier_id),
                        parameter = "\n    ".join(map(lambda x: x.setRequest(), self.request_params)),
                        parameter_length = str(sum(map(lambda x: x.getLength(), self.request_params))) + "+".join([""] + filter(lambda x: x != None, map(lambda x: x.getDynamicLength(), self.request_params))),
                        timeout = self.timeout,
                        parse_response = self.parse_method_call("response"))
ApiCall.sync_method = sync_method

def callback_selector(self, call_type, long_message=False):
    if call_type == "response":
        type_const = "DS485_CONTAINER_RESPONSE"
        parameters = self.response_params
    elif call_type == "request":
        type_const = "DS485_CONTAINER_REQUEST"
        parameters = self.request_params
    else:
        type_const = "DS485_CONTAINER_EVENT"
        parameters = self.response_params
    
    s = Template("""    if (MessageType == $message_type && MessageId == ${message_id} && ModifierId == ${modifier_id})
    {
        ${param_declarations}
        error = ${call_parse_method};
        ((${name}_${call_type}_callback_t)callback_entry->callback)(error, callback_entry->arg, ${container}->sourceId, ${container}->destinationId${parameter_call});

        return 0;
    }""")
    return s.substitute(message_type = type_const,
                        message_id = str(self.message_id),
                        modifier_id = str(self.modifier_id),
                        name = self.name,
                        call_type = call_type,
                        container = "container" if not long_message else "packet",
                        param_declarations = "\n        ".join(["uint8_t error;"] + map(lambda x: x.declaration() + ";", parameters)),
                        call_parse_method = self.parse_method_call_by_reference(call_type, "container" if not long_message else "packet"),
                        parameter_call = "".join(map(lambda x: ", " + x.call(), parameters)))
ApiCall.callback_selector = callback_selector

def parse_method_call_by_reference(self, call_type, container):
    if call_type == "request":
        parameters = self.request_params
    else:
        parameters = self.response_params
    
    out = self.name + "_parse_" + call_type + "(" + container
    if parameters:
        out += ", " + ", ".join(map(lambda x: x.call_as_object_by_reference(), parameters))
    out += ")"
    return out
ApiCall.parse_method_call_by_reference = parse_method_call_by_reference

def documentationStart(self):
    s = Template("""/**
 * @defgroup $name $name
 *
 * $documentation
 * @{
 */
 """)
    return s.substitute(name = self.name,
                        documentation = self.description)
CallGroup.documentationStart = documentationStart

def documentationEnd(self):
    return """/** @} */
"""
CallGroup.documentationEnd = documentationEnd

def header(self):
    s = ""
    s += self.documentationStart()
    for call in self.calls:
        s += call.documentation_call() + "\n"
        s += call.call_signature(header=True) + ";\n\n"
        s += call.callback_documentation("response") + "\n"
        s += call.callback_declaration("response") + "\n\n"
        s += call.callback_documentation("request") + "\n"
        s += call.callback_declaration("request") + "\n\n"
        
    s += "\n\n"
    for event in self.events:
        s += event.callback_documentation("event") + "\n"
        s += event.callback_declaration("event") + "\n\n"
    s += self.documentationEnd()
    return s
CallGroup.header = header

def implementation(self):
    s = ""
    for call in self.calls:
        s += call.parse_method_signature("response") + ";\n"
        s += call.parse_method_signature("request") + ";\n"
    for event in self.events:
        s += event.parse_method_signature("event") + ";\n"
    
    s += "\n"
    for call in self.calls:
        s += call.parse_method("response") + "\n\n"
        s += call.parse_method("request") + "\n\n"
        s += call.sync_method() + "\n\n"
        
    for event in self.events:
        s += event.parse_method("event") + "\n\n"
    return s
CallGroup.implementation = implementation

def callback_selector(self):
    s = ""
    for call in self.calls:
        if not call.long_message:
            s += call.callback_selector("response") + "\n\n"
            s += call.callback_selector("request") + "\n\n"
    for event in self.events:
        if not event.long_message:
            s += event.callback_selector("event") + "\n\n"
    return s
CallGroup.callback_selector = callback_selector

def callback_long_selector(self):
    s = ""
    for call in self.calls:
        if call.long_message:
            s += call.callback_selector("response", long_message=True) + "\n\n"
            s += call.callback_selector("request", long_message=True) + "\n\n"
    for event in self.events:
        if event.long_message:
            s += event.callback_selector("event", long_message=True) + "\n\n"
    return s
CallGroup.callback_long_selector = callback_long_selector
        
class CGenerator:
    def __init__(self, license, groups, version):
        self.groups = groups
        self.license = license
        self.version = version
        
    def header(self, main_page_doc, header_file):
        header = open(header_file, 'w')
        header_str = self.license
        header_str += "/** \mainpage dSM-API Documentation\n"
        header_str += " * " + main_page_doc
        header_str += """\n *\n *
 * NOTE: The libdsm-api and it's scripting bindings are unsupported interfaces and 
 * subject to change without further notice.
"""
        header_str += " */\n"
        header_str += """/**
 * @file dsm-api.h
 *
 * This is the dSM-API declaration.
 *
 * General remark: If any return parameters are of no interest to the caller
 *                 NULL can be passed instead of a variable reference.
 */

#ifndef DSM_API_H
#define DSM_API_H
#include <digitalSTROM/ds485-client-interface.h>

#ifdef __cplusplus
extern "C" {
#endif

"""
        header_str += "#define DSM_API_VERSION %s\n" % self.version
        header_str += """
/**
 * @defgroup Library Library
 *
 * This group of API calls addresses the management of the dS485-Stack, connection to the ds485d daemon and the callback infrastructure.
 *
 * @{
 */

/**
 * Type of the library handle.
 */
typedef void *DsmApiHandle_t;

/**
 * Callback function signature for the BusChange callback.
 */
typedef void (*BusChange_callback_t)(void *arg, dsuid_t *id, int flag);

/**
 * Callback function signature for the BusState callback.
 */
typedef void (*LinkState_callback_t)(void *arg, bool state);

/**
 * Callback function signature for the BusState callback.
 */
typedef void (*DsmApiBlocking_callback_t)(void *arg);

/**
 * Data structure for dSM-API callback registration
 */
typedef struct {
    void *function;
    void *arg;
} DsmApiCallback_t;

/**
 * Get the API version of the library.
 *
 * This function returns the API version as two bytes. The first byte is the MAJOR version,
 * the second byte is the MINOR version. This version corresponds to the ApiVersion parameter read
 * from the dSM with #dSMInfo call. Differing MAJOR versions are binary incompatible. Higher MINOR
 * versions contain additional functionality not previously available.
 *
 * @return API version
 */
int DsmApiLibraryVersion(void);

/**
 * Initialize the internal memory structures of the library.
 *
 * @return a handle to the library instance
 */
DsmApiHandle_t DsmApiInitialize(void);

/**
 * Open a new connection to the dS485 bus.
 * 
 * @param [in] handle Handle to the library instance
 * @param [in] connection string containing the connection method: 
 *                        - serial:///dev/ttyUSB0 (only when linked to the GPL version of the ds485-client library)
 *                        - tcp://localhost:8442
 * @param [in] flags
 * @return zero on success, negative on failure
 */
int DsmApiOpen(DsmApiHandle_t handle, const char *connection, int flags);

/**
 * Get the dSUID of the daeomon/socket interface.
 *
 * @param [in] handle Handle to the library instance
 * @param [out] own dSUID
 * @return 0 on success; negative on failure
 */
int DsmApiGetOwnDSUID(DsmApiHandle_t handle, dsuid_t *dsuid);

/**
 * Request a list of the current bus members.
 * 
 * @param [in] handle Handle to the library instance
 * @param [out] device_list List of devices. Preallocate a buffer of size device_list_size * sizeof(dsuid_t)
 * @param [in] device_list_size device_list has space for this many devices.
 * @return the number of devices in device_list; negative if an error occured
 */
int DsmApiGetBusMembers(DsmApiHandle_t handle, dsuid_t *device_list, int device_list_size);

/**
 * Request a list of the current bus members depending on their type.
 *
 * @param [in] handle Handle to the library instance
 * @param [out] device_list List of devices. Preallocate a buffer of size device_list_size * sizeof(dsuid_t)
 * @param [in] device_list_size device_list has space for this many devices.
 * @param [in] types type filter for device list.
 * @return the number of devices in device_list; negative if an error occured
 */
int DsmApiGetBusMembersPerType(DsmApiHandle_t handle, dsuid_t *device_list, int device_list_size, unsigned int types);

/**
 * Get the current state of the dS485 bus.
 *
 * @param [in] handle Handle to the library instance
 * @param [out] state returns the bus state
 * @return 0 on success; negative on failure
 */
int DsmApiGetLinkState(DsmApiHandle_t handle, bool *state);

/**
 * Restart the dS485 bus.
 *
 * @param [in] handle Handle to the library instance
 * @return 0 on success; negative on failure
 */
int DsmApiRestartBus(DsmApiHandle_t handle);

/**
 * Close the connection to the dS485 bus.
 *
 * @param [in] handle Handle to the library instance
 * @return 0 on success; negative on failure
 */
int DsmApiClose(DsmApiHandle_t handle);

/**
 * Free instance memory.
 *
 * @param [in] handle Handle to the library instance
 * @return 0 on success; negative on failure
 */
int DsmApiCleanup(DsmApiHandle_t handle);

/**
 * Register callback functions for dsm-api asynchronous responses, requests and events.
 *
 * Register NULL to delete a callback.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] MessageType Register callback function for this MessageType
 * @param [in] MessageId Register callback function for this MessageId
 * @param [in] ModifierId Register the callback function for this ModifierId
 * @param [in] callback a struct containing a callback_function and a user defined argument pointer
 * @param [out] oldcallback a struct containing the previously registered callback upon return
 * @return ERROR_OK when ok
 */
int DsmApiSetCallback(DsmApiHandle_t handle, container_type_t MessageType, uint8_t MessageId, uint8_t ModifierId, DsmApiCallback_t *const callback, DsmApiCallback_t *oldcallback);

/**
 * Register a callback function for BusState events.
 *
 * Register NULL to delete the callback.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] callback a struct containing a callback_function and a user defined argument pointer
 * @param [out] oldcallback a struct containing the previously registered callback upon return
 * @return ERROR_OK when ok
 */
int DsmApiSetLinkStateCallback(DsmApiHandle_t handle, DsmApiCallback_t *const callback, DsmApiCallback_t *oldcallback);

/**
 * Register a callback function for BusChange events.
 *
 * Register NULL to delete the callback.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] callback a struct containing a callback_function and a user defined argument pointer
 * @param [out] oldcallback a struct containing the previously registered callback upon return
 * @return ERROR_OK when ok
 */
int DsmApiSetBusChangeCallback(DsmApiHandle_t handle, DsmApiCallback_t *const callback, DsmApiCallback_t *oldcallback);

/**
 * Register a diagnostic callback function called before blocking operations.
 *
 * Register NULL to delete the callback.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] callback a struct containing a callback_function and a user defined argument pointer
 * @param [out] oldcallback a struct containing the previously registered callback upon return
 * @return ERROR_OK when ok
 */
int DsmApiSetBlockingCallback(DsmApiHandle_t handle, DsmApiCallback_t *const callback, DsmApiCallback_t *oldcallback);

/** @} */

/**
 * @defgroup Convenience Convenience Functions
 *
 * This is a group of convenience functions used to simplify the interaction with digitalSTROM(r) devices.
 * @{
 */

/**
 * Convert a combination of product properties and serial number to a unique dSUID.
 *
 * @param [in] vendorId VendorId of the device
 * @param [in] productId ProductId of the device
 * @param [in] groupId Standard-Function of the device
 * @param [in] revisionId Firmare RevisionId of the device
 * @param [in] serialNumber Serial number of the device
 * @param [out] dsuid Contains the dSUID of the device.
 * @return 0 on success, ERROR_DSUID_CONVERSION if product is unknown, dSUID is 0 and contains only serial number
 */
int DsmApiGetDeviceDSUID(uint16_t vendorId, uint16_t productId, uint8_t groupId, uint16_t revisionId, uint32_t serialNumber,
    dsuid_t *dsuid);

/**
 * Convert a IEEE address into a unique dSUID.
 *
 * @param [in] mac MAC Address of the device
 * @param [out] dsuid Contains the dSUID of the device.
 * @return 0, if IEEE prefix and device is known, else ERROR_DSUID_CONVERSION
 */
int DsmApiGetEthernetDSUID(const char *mac, dsuid_t *dsuid);

/**
 * Expand a combination of product properties to a universal description and GTIN number.
 *
 * @param [in] vendorId VendorId of the device
 * @param [in] productId ProductId of the device
 * @param [in] groupId Standard-Function of the device
 * @param [in] revisionId Firmare RevisionId of the device
 * @param [out] displayName Contains the display name of the product
 * @param [out] hardwareInfo Contains the hardware info string
 * @param [out] gtin Contains the GTIN string
 * @return 0 on success, -1 if product is unknown
 */
int DsmApiGetDeviceDescription(uint16_t vendorId, uint16_t productId, uint8_t groupId, uint16_t revisionId,
    char **displayName, char **hardwareInfo, char **gtin);

/**
 * Check if a given dSUID belongs to a dSM.
 *
 * @param [in] dSMdSUID dSUID of the device in question
 * @return 0: device is no dSM, 1: device is a dSM
 */
int DsmApiIsdSM(dsuid_t dSMdSUID);

/** @} */

/** @ingroup Device
 * Requests statistics data of a device. Synchronous implementation of #DeviceStatisticsRequest_get.
 *
 * ParamOffset 0xFE (Reset) and 0xFF (All values) are not allowed, as they would result in no or multiple response events.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId 
 * @param [in] ParamOffset 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Statistics value (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceStatisticsRequest_get_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ParamOffset, int timeout, uint16_t *Value);

/** @ingroup Device
 * Read a configuration parameter of a specific digitalSTROM device. Synchronous implementation of #DeviceConfig_get.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId 
 * @param [in] ConfigClass 
 * @param [in] ConfigIndex 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Parameter value (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceConfig_get_sync_8(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint8_t *Value);

/** @ingroup Device
 * Read two configuration parameters of a specific digitalSTROM device simultaneously. Synchronous implementation of #DeviceConfig_get.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId 
 * @param [in] ConfigClass 
 * @param [in] ConfigIndex 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Parameter values low byte: ConfigIndex & 0xFE; high byte: ConfigIndex & 0xFE + 1 (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceConfig_get_sync_16(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint16_t *Value);

/** @ingroup Device
 * Set a configuration parameter of a specific digitalSTROM device. Synchronous implementation of #DeviceConfig_set.
 *
 * Caution: This function doesn't guarantee that the device received the Value. On success, it just means that the dSM could sent the Value in given time.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] ConfigClass
 * @param [in] ConfigIndex
 * @param [in] Value
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceConfig_set_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, uint8_t Value, int timeout);

/** @ingroup Device
 * Set a zone of a specific digitalSTROM device. Synchronous implementation of #DeviceProperties_set_zone.
 *
 * Caution: This function doesn't guarantee that the device received the value. On success, it just means that the dSM could sent the ZoneId in given time.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId
 * @param [in] ZoneId
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceProperties_set_zone_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint16_t ZoneId, int timeout);

/** @ingroup Device
 * Set a button output channel of a specific digitalSTROM device. Synchronous implementation of #DeviceProperties_set_button_set_output_channel.
 *
 * Caution: This function doesn't guarantee that the device received the value. On success, it just means that the dSM could sent the ChannelId in given time.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId
 * @param [in] ChannelId
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceProperties_set_button_set_output_channel_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ChannelId, int timeout);

/** @ingroup Device
 * Set a button global group of a specific digitalSTROM device. Synchronous implementation of #DeviceProperties_set_button_active_group.
 *
 * Caution: This function doesn't guarantee that the device received the value. On success, it just means that the dSM could sent the GroupId in given time.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId
 * @param [in] GroupId
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceProperties_set_button_active_group_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t GroupId, int timeout);

/** @ingroup Device
 * Add a group to a specific digitalSTROM device. Synchronous implementation of #DeviceGroupMembershipModify_add.
 *
 * Caution: This function doesn't guarantee that the device received the value. On success, it just means that the dSM could sent the GroupId in given time.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId
 * @param [in] GroupId
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceGroupMembershipModify_add_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t GroupId, int timeout);

/** @ingroup Device
 * Read the current value of an output channel of a digitalSTROM device. Synchronous implementation of #DeviceOPCConfig_get_current.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId 
 * @param [in] ChannelId 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Parameter values (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceOPCConfig_get_current_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ChannelId, int timeout, uint16_t *Value);

/** @ingroup Device
 * Read the scene value of an output channel of a digitalSTROM device. Synchronous implementation of #DeviceOPCConfig_get_scene.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId 
 * @param [in] ChannelId 
 * @param [in] SceneId 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Parameter values (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceOPCConfig_get_scene_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ChannelId, uint8_t SceneId, int timeout, uint16_t *Value);

/** @ingroup Device
 * Read the scene config value for a scene of a digitalSTROM device
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId
 * @param [in] SceneId
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Parameter values (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceOPCConfig_get_scene_config_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t SceneId, int timeout, uint16_t *Value);

/** @ingroup Device
 * Read don't care flags bitfield for a scene of a digitalSTROM device. Synchronous implementation of #DeviceOPCConfig_get_dc_flags.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId 
 * @param [in] SceneId 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] Value Parameter values (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceOPCConfig_get_dc_flags_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t SceneId, int timeout, uint16_t *Value);

/** @ingroup Device
 * Query transmission quality. Synchronous implementation of #TestTransmissionQuality_get.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId 
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] DownstreamQuality Downstream quality (only valid if return == 0)
 * @param [out] UpstreamQuality Upstream quality (only valid if return == 0)
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int TestTransmissionQuality_get_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, int timeout, uint8_t *DownstreamQuality, uint16_t *UpstreamQuality);

/** @ingroup Device
 * Access device sensors. Get the value of sensor SensorIndex. Synchronous implementation of #DeviceSensor_get_value.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId DeviceId
 * @param [in] SensorIndex Index of the requested sensor on the device.
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] SensorValue Current value of the sensor (12 bit).
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceSensor_get_value_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t SensorIndex, int timeout, uint16_t *SensorValue);

/** @ingroup Device
 * Access device sensors with extended data. Get the value of sensor SensorIndex as double, age and context data. Synchronous implementation of #DeviceSensor_get_value_extended.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId DeviceId
 * @param [in] SensorIndex Index of the requested sensor on the device.
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] SensorValue Current value of the sensor (64 bit).
 * @param [out] valueAge Current age of the value in milliseconds (32 bit).
 * @param [out] contextId Additional context data related to the measurement (32 bit).
 * @param [out] contextMsg Additional context string related to the measurement (64 byte).
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceSensor_get_value_extended_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t SensorIndex, int timeout, double *SensorValue, uint32_t *valueAge, uint32_t *contextId, uint8_t *contextMsg);

/** @ingroup dSLink
 * Send UDI commands to a single dSLink device. UDI device stays selected after the call execution. Used when multiple bytes have to be sent. Synchronous implementation of #DeviceDslinkSend_not_last_byte_with_answer.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId Select only this device.
 * @param [in] Value Value to send to devices.
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] ReturnValue Answer from the device
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceDslinkSend_not_last_byte_with_answer_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t Value, int timeout, uint8_t *ReturnValue);

/** @ingroup dSLink
 * Send UDI commands to a single dSLink device. UDI device is de-selected after the call. Used only on the last access. Synchronous implementation of #DeviceDslinkSend_last_byte_with_answer.
 *
 * @param [in] handle Handle to the library instance
 * @param [in] dSMdSUID Destination of the call
 * @param [in] DeviceId Select only this device.
 * @param [in] Value Value to send to devices.
 * @param [in] timeout time-out of the synchronous call (in Seconds)
 * @param [out] ReturnValue Answer from the device
 * @return 0: success (return parameters are valid); <0: Library error; >0: dSM error code
 */
int DeviceDslinkSend_last_byte_with_answer_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t Value, int timeout, uint8_t *ReturnValue);
"""
        for group in self.groups:
            header_str += group.header()
            
        header_str += """#ifdef __cplusplus
}
#endif

#endif
"""
        header.write(header_str.encode('utf-8'))
        header.close()
        
    def implementation(self, impl_file):
        source = open(impl_file, 'w')
        source_str = self.license
        source_str += """#include "dsm-api.h"
#include "dsm-api-const.h"
#include <digitalSTROM/ds485-client-interface.h>
#include "dsm-macros.h"

#include "dsm-api-static.c"
"""
        for group in self.groups:
            source_str += group.implementation()
    
        source_str += """static int dsm_api_ds485_message_callback(void *handle, const ds485_container_t *container)
{
    uint8_t MessageId;
    uint8_t ModifierId;
    container_type_t MessageType;
    uint32_t hash_key;
    struct callback_item *callback_entry = NULL;
    
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }
    
    if (container == NULL)
    {
        return -1;
    }

    MessageId = GetCommand(*container);
    ModifierId = GetModifier(*container);
    MessageType = container->containerType;
    hash_key = (MessageType << 16) | (MessageId << 8) | (ModifierId);

    pthread_mutex_lock(&data->mutex_callback);
    HASH_FIND_INT(data->callback_list, &hash_key, callback_entry);
    pthread_mutex_unlock(&data->mutex_callback);
    if (callback_entry == NULL)
    {
        return -1;
    }
    
    if (callback_entry->callback == NULL)
    {
        return -1;
    }

"""
        for group in self.groups:
            source_str += group.callback_selector()
        source_str += """    return -1;
}
"""

        source_str += """
static int dsm_api_ds485_netlib_packet_callback(void *handle, const ds485n_packet_t *packet)
{
    uint8_t MessageId;
    uint8_t ModifierId;
    container_type_t MessageType;
    uint32_t hash_key;
    struct callback_item *callback_entry = NULL;

    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    if (packet == NULL)
    {
        return -1;
    }

    MessageId = GetCommand(*packet);
    ModifierId = GetModifier(*packet);
    MessageType = packet->containerType;
    hash_key = (MessageType << 16) | (MessageId << 8) | (ModifierId);

    pthread_mutex_lock(&data->mutex_callback);
    HASH_FIND_INT(data->callback_list, &hash_key, callback_entry);
    pthread_mutex_unlock(&data->mutex_callback);
    if (callback_entry == NULL)
    {
        return -1;
    }

    if (callback_entry->callback == NULL)
    {
        return -1;
    }

"""
        for group in self.groups:
            source_str += group.callback_long_selector()
        source_str += """    return -1;
}
"""
        source.write(source_str.encode('utf-8'))
        source.close()
    
    def const_header(self, const_header_file, message_ids, error_list,
                     selectors_bus_type, values_bus_type, length_enum):
        const = open(const_header_file, 'w')
        const_str = self.license
        const_str += """#ifndef DSM_API_CONST_H
#define DSM_API_CONST_H

#ifdef __cplusplus
extern "C" {
#endif

"""
        
        const_str += "enum DsmApiMsgId {\n"
        for i in message_ids:
            const_str += "    " + i.get_enum_decl() + ",\n"
        const_str += "};\n"
        
        for i in message_ids:
            const_str += i.get_modifier_enum() + "\n\n"
        
        if (len(error_list) > 0):
            const_str += "enum DsmApiErrorCodes {\n"
            for key, value in error_list.iteritems():
                const_str += "    " + value.getEnumEntry() + "\n"
            const_str += "};\n"

        if selectors_bus_type:
            const_str += "typedef enum {\n"
            for value in selectors_bus_type:
                const_str += "    " + value.getEnumEntry() + "\n"
            const_str += "} Selector_BusMembersType_t;\n\n"

        if values_bus_type:
            const_str += "typedef enum {\n"
            for value in values_bus_type:
                const_str += "    " + value.getEnumEntry() + "\n"
            const_str += "} BusMemberDevice_t;\n\n"
            
        if (len(length_enum) > 0):
            const_str += "enum DsmApiParameterLength {\n"
            for key, value in length_enum.iteritems():
                const_str += "    " + key + " = " + str(value) + ",\n"
            const_str += "};\n\n"
        
        const_str += """#ifdef __cplusplus
}
#endif

#endif /* DSM_API_CONST_H */
"""
            
        const.write(const_str.encode('utf-8'))
        const.close()
        
